from enum import Enum

from kognic.io.model.base_serializer import BaseSerializer


class CalibrationType(str, Enum):
    PINHOLE = "pinhole"
    FISHEYE = "fisheye"
    KANNALA = "kannala"
    LIDAR = "lidar"
    PRINCIPALPOINTDIST = "principal_point_distortion"
    FUSEDCYLINDRICAL = "fused_cylindrical"


class RotationQuaternion(BaseSerializer):
    w: float
    x: float
    y: float
    z: float


class Position(BaseSerializer):
    x: float
    y: float
    z: float


class BaseCalibration(BaseSerializer):
    position: Position
    rotation_quaternion: RotationQuaternion



from kognic.io.model.input.aggregated_lidars_and_cameras_seq.aggregated_lidars_and_cameras_seq import AggregatedLidarsAndCamerasSequence
from kognic.io.model.input.aggregated_lidars_and_cameras_seq.frame import Frame


from typing import Optional, List

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration.camera.common import BaseCameraCalibration
from kognic.io.model.calibration.common import CalibrationType


class PrincipalPointDistortionCoefficients(BaseSerializer):
    k1: float
    k2: float


class PrincipalPoint(BaseSerializer):
    x: float
    y: float


class DistortionCenter(BaseSerializer):
    x: float
    y: float


class LensProjectionCoefficients(BaseSerializer):
    c1: float
    c2: float
    c3: float
    c4: float
    c5: float
    c6: float

    def get_coefficients(self) -> List[float]:
        return [self.c1, self.c2, self.c3, self.c4, self.c5, self.c6]


class PrincipalPointDistortionCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.PRINCIPALPOINTDIST.value
    principal_point_distortion_coefficients: PrincipalPointDistortionCoefficients
    distortion_center: DistortionCenter
    principal_point: PrincipalPoint
    lens_projection_coefficients: Optional[LensProjectionCoefficients]



import urllib
from typing import Callable
from typing import Optional, Generator

from kognic.base_clients.cloud_storage import FileResourceClient
from kognic.base_clients.http_client import HttpClient
from kognic.base_clients.models import CursorIdType


class IOResource:

    def __init__(self, client: HttpClient, file_client: FileResourceClient):
        super().__init__()
        self._client = client
        self._file_client = file_client

    def _paginate_get(self, endpoint: str, **kwargs) -> Generator[dict, None, None]:
        yield from self._paginate(self._client.get, endpoint, **kwargs)

    def _paginate_post(self, endpoint: str, **kwargs) -> Generator[dict, None, None]:
        yield from self._paginate(self._client.post, endpoint, **kwargs)

    def _paginate(
        self,
        api_method: Callable,
        endpoint: str,
        next_cursor_id: Optional[CursorIdType] = None,
        **kwargs,
    ) -> Generator[dict, None, None]:
        """
        Paginates through result pages recursively.

        :param api_method: api method to call in the form of a function
        :param endpoint: endpoint to send request to
        :param next_cursor_id: next cursor id to call. Will be appended with the key 'cursorId' to the endpoint
        :param **kwargs: Passed to api method
        """

        next_page_url = urllib.parse.urljoin(endpoint, f"?cursorId={next_cursor_id}") if next_cursor_id is not None else endpoint
        page = api_method(next_page_url, **kwargs)
        for item in page.data:
            yield item

        if page.metadata.next_cursor_id is not None:
            yield from self._paginate(api_method, endpoint, page.metadata.next_cursor_id, **kwargs)



from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field

from kognic.io.model.base_serializer import BaseSerializer


class InputStatus(str, Enum):
    Pending = "pending"
    Processing = "processing"
    Created = "created"
    Failed = "failed"
    InvalidatedBadContent = "invalidated:broken-input"
    InvalidatedSlamRerun = "invalidated:slam-rerun"
    InvalidatedDuplicate = "invalidated:duplicate"
    InvalidatedIncorrectlyCreated = "invalidated:incorrectly-created"


class Input(BaseSerializer):
    uuid: str = Field(alias='internalId')
    external_id: str
    batch: str = Field(alias='batchId')
    input_type: str
    status: InputStatus
    created: datetime
    annotation_types: List[str]
    calibration_id: Optional[str]
    view_link: Optional[str]
    error_message: Optional[str]



import logging
from typing import List, Optional

from kognic.io.model.input.input_entry import Input
from kognic.io.model.input.invalidated_reason_input import InvalidatedReasonInput
from kognic.io.resources.abstract import IOResource
from kognic.io.util import filter_none

log = logging.getLogger(__name__)


class InputResource(IOResource):
    """
    Resource exposing Kognic Inputs
    """

    def invalidate_inputs(self, input_uuids: List[str], invalidated_reason: InvalidatedReasonInput) -> None:
        """
        Invalidates inputs, and removes them from all input lists

        :param input_uuids: The input uuids to invalidate
        :param invalidated_reason: An Enum describing why inputs were invalidated
        """
        invalidated_json = dict(inputIds=input_uuids, invalidatedReason=invalidated_reason)
        self._client.post("v1/inputs/actions/invalidate", json=invalidated_json, discard_response=True)

    def get_inputs(
        self,
        project: str,
        batch: Optional[str] = None,
        include_invalidated: bool = False,
        external_ids: Optional[List[str]] = None,
    ) -> List[Input]:
        """
        Gets inputs for project, with option to filter for invalidated inputs

        :param project: Project (identifier) to filter
        :param batch: Batch (identifier) to filter
        :param invalidated: Returns invalidated inputs if True, otherwise valid inputs
        :param external_id: External ID to filter input on
        :param paginate: Only returns the first page if set to false
        :return List: List of Inputs
        """

        payload = {
            "project": project,
            "batch": batch,
            "includeInvalidated": include_invalidated,
            "externalIds": external_ids,
        }

        endpoint = "v1/inputs/fetch"

        inputs = list()
        for js in self._paginate_post(endpoint, json=filter_none(payload)):
            inputs.append(Input.from_json(js))
        return inputs

    def add_annotation_type(self, input_uuid: str, annotation_type: str) -> None:
        """
        Adds annotation-type to the input, which informs the Kognic Platform
        to produce a corresponding annotation for the annotation-type. Only
        possible if the annotation-type is available in the corresponding batch
        of the input (use method `get_annotation_types` to check).
        """

        self._client.post(f"v1/inputs/{input_uuid}/actions/add-annotation-type/{annotation_type}", discard_response=True)

    def remove_annotation_types(self, input_uuid: str, annotation_types: List[str]) -> None:
        """
        Removes annotation types from the input, which informs the Kognic Platform
        that a corresponding annotation should not be produced for the annotation types. Only
        possible if the annotation type is available for the input (use method `get_inputs_by_uuids` to check).
        Note: If multiple annotation types are configured to be annotated at the same time, i.e. on the same request,
        all of these annotation types need to be provided.
        """
        body = dict(annotationTypes=annotation_types)
        self._client.post(f"v1/inputs/{input_uuid}/annotation-types/actions/remove", json=body, discard_response=True)

    def get_inputs_by_uuids(self, input_uuids: List[str]) -> List[Input]:
        """
        Gets inputs using input uuids

        :param input_uuids: A UUID to filter inputs on
        :return List: List of Inputs
        """

        body = dict(uuids=input_uuids)
        json_resp = self._client.post("v1/inputs/query", json=body)
        return [Input.from_json(js) for js in json_resp]



from typing import Optional, Dict

from pydantic import BaseModel, root_validator

allowed_types = (int, float, str, bool)
allowed_types_pretty = [t.__name__ for t in allowed_types]


class MetaDataBase(BaseModel):
    """
    Base class for MetaData which only allows flat key-value pairs.
    """

    class Config:
        extra = "allow"

    @root_validator
    def validate_flat_types(cls, values):
        reserved_fields = cls.schema()['properties']
        for key, val in values.items():
            if key in reserved_fields:
                continue

            if not isinstance(val, allowed_types):
                raise ValueError(
                    f"Illegal Metadata value type for field '{key}'. Got '{type(val).__name__}', \
                    but expected one of {allowed_types_pretty}"
                )

        return values


class MetaData(MetaDataBase):
    """
    Container for metadata belonging to the input.
    The attributes of this class are reserved keywords
    with assiciated functionality in the Kognic platform.

    Attributes:
        region: A string indicating the region the data was collected in. 
            If there are annotation instructions associated with a specific
            region, talk to your contact at Kognic to sync what should be
            specified in the region attribute.
    """
    region: Optional[str]


class FrameMetaData(MetaDataBase):
    pass



"""Client for communicating with the Kognic platform."""
import logging
from typing import Optional

from kognic.auth import DEFAULT_HOST as DEFAULT_AUTH_HOST

from kognic.base_clients.cloud_storage import FileResourceClient
from kognic.base_clients.http_client import HttpClient

from kognic.io.resources.annotation.annotation import AnnotationResource
from kognic.io.resources.calibration.calibration import CalibrationResource
from kognic.io.resources.input.cameras import Cameras
from kognic.io.resources.input.cameras_sequence import CamerasSequence
from kognic.io.resources.input.input import InputResource
from kognic.io.resources.input.lidars import Lidars
from kognic.io.resources.input.lidars_and_cameras import LidarsAndCameras
from kognic.io.resources.input.lidars_and_cameras_sequence import LidarsAndCamerasSequence
from kognic.io.resources.input.aggregated_lidars_and_cameras_seq import AggregatedLidarsAndCamerasSequence
from kognic.io.resources.input.lidars_sequence import LidarsSequence
from kognic.io.resources.input.pre_annotation import PreAnnotationResource
from kognic.io.resources.project.project import ProjectResource

DEFAULT_HOST = "https://input.app.kognic.com"

log = logging.getLogger(__name__)


class KognicIOClient:
    """Client to work upload and retrieve data from the Kognic platform"""

    def __init__(
        self,
        *,
        auth=None,
        host: str = DEFAULT_HOST,
        auth_host: str = DEFAULT_AUTH_HOST,
        client_organization_id: Optional[int] = None,
        max_retry_attempts: int = 23,
        max_retry_wait_time: int = 60,
        timeout: int = 60
    ):
        """
        :param auth: auth credentials, see https://developers.kognic.com/docs/kognic-auth
        :param host: override for api url
        :param auth_host: override for authentication url
        :param client_organization_id: Overrides your users organization id. Only works with an Kognic user.
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        :param timeout: Max time to wait for response from server.
        """

        self._client = HttpClient(
            auth=auth,
            host=host,
            auth_host=auth_host,
            client_organization_id=client_organization_id,
            timeout=timeout,
        )
        self._file_client = FileResourceClient(
            max_retry_attempts=max_retry_attempts,
            max_retry_wait_time=max_retry_wait_time,
            timeout=timeout,
        )

        self.calibration = CalibrationResource(self._client, self._file_client)
        self.project = ProjectResource(self._client, self._file_client)
        self.annotation = AnnotationResource(self._client, self._file_client)
        self.input = InputResource(self._client, self._file_client)
        self.pre_annotation = PreAnnotationResource(self._client, self._file_client)

        self.lidar_and_cameras = LidarsAndCameras(self._client, self._file_client)
        self.lidars_and_cameras_sequence = LidarsAndCamerasSequence(self._client, self._file_client)
        self.cameras = Cameras(self._client, self._file_client)
        self.cameras_sequence = CamerasSequence(self._client, self._file_client)
        self.lidars = Lidars(self._client, self._file_client)
        self.lidars_sequence = LidarsSequence(self._client, self._file_client)
        self.aggregated_lidars_and_cameras_seq = AggregatedLidarsAndCamerasSequence(self._client, self._file_client)



from typing import Dict

from humps import camelize
from pydantic import BaseModel


class BaseSerializer(BaseModel):

    @classmethod
    def from_json(cls, js: Dict):
        return cls.parse_obj(js)

    def to_dict(self, by_alias=True) -> Dict:
        return self.dict(exclude_none=True, by_alias=by_alias)

    class Config:
        alias_generator = camelize
        orm_mode = True
        allow_population_by_field_name = True



from typing import List, Optional, Mapping

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.ego.utils import UnixTimestampNs
from kognic.io.model.input.resources.image import Image
from kognic.io.model.input.resources.point_cloud import PointCloud


class Frame(BaseSerializer):
    point_clouds: List[PointCloud]
    images: List[Image]
    unix_timestamp: Optional[UnixTimestampNs] = None

    @property
    def resources(self) -> Mapping[str, Image]:
        img_mappings = {i.resource_id: i for i in self.images}
        pc_mappings = {p.resource_id: p for p in self.point_clouds}
        return {**img_mappings, **pc_mappings}



from typing import Optional

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration.camera.common import BaseCameraCalibration
from kognic.io.model.calibration.common import CalibrationType


class CutAngles(BaseSerializer):
    upper: float
    lower: float


class FusedCylindricalCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.FUSEDCYLINDRICAL.value
    cut_angles_degree: CutAngles
    vertical_fov_degree: Optional[float] = 72.5
    horizontal_fov_degree: Optional[float] = 93



import logging
import time
from typing import Optional, List, Type, Generator

from kognic.io.client import KognicIOClient
from kognic.io.model.input import LidarsAndCamerasSequence, Cameras, CamerasSequence, LidarsAndCameras, \
    AggregatedLidarsAndCamerasSequence, InputStatus
from kognic.io.model.input.feature_flags import FeatureFlags
from kognic.io.resources.abstract import CreateableIOResource
from kognic.io.tools.model import SceneWithPreAnnotation, InputCreationResult, InputCreationError, InputCreationStage, InputCreationStatus, \
    Scene, SceneUuid
from kognic.openlabel.models import OpenLabelAnnotation
from requests import HTTPError
from tqdm import tqdm

logger = logging.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 100
CONSERVATIVE_CHUNK_SIZE = 20


def create_inputs(
    client: KognicIOClient,
    scenes_with_pre_annotations: List[SceneWithPreAnnotation],
    project: str,
    dryrun: Optional[bool] = False,
    feature_flags: Optional[FeatureFlags] = None,
    wait_timeout: float = 1800,  # this will depend heavily on the scene size and the number of scenes
    sleep_time: float = 1,
    **kwargs
) -> Generator[InputCreationResult, None, None]:
    """
    Create multiple inputs, optionally with pre-annotations. This consists of the following steps
    1. Start scene jobs
    2. As scene jobs complete
        2.1. A pre-annotation is optionally created
        2.2. An input is created for the scene

    :param client: KognicIOClient
    :param scenes_with_pre_annotations: List of scenes and pre-annotations to create. Note that the scenes must all be of the same type.
    :param project: project to add input to
    :param dryrun: If True the files/metadata will be validated but no scene job will be created.
    :param feature_flags: Optional set of feature flags to control the scene creation process.
    :param wait_timeout: Time in seconds to wait for scene jobs to complete.
    :param sleep_time: Time in seconds to sleep between checking scene job status.
    :param kwargs: Passed to the `create_from_scene` method. For example batch and annotation_types
    :returns InputCreationResult: Contains information about the creation process. For example, whether it succeeded or failed
    """

    scenes = [swp.scene for swp in scenes_with_pre_annotations]
    resource = _get_creatable_resource(client, scenes)
    scene_uuids = _start_scene_jobs(resource, scenes, dryrun=dryrun, feature_flags=feature_flags)

    if dryrun:  # Don't do anything more if dry-run
        return

    scenes_map = {uuid: swp for uuid, swp in zip(scene_uuids, scenes_with_pre_annotations)}
    for input_result in _wait_for_scene_jobs(client, scene_uuids, wait_timeout, sleep_time):
        if input_result.status == InputCreationStatus.CREATED:
            pre_annotation = scenes_map[input_result.scene_uuid].pre_annotation
            if pre_annotation:
                pre_anno_result = _create_pre_annotation(client, input_result.scene_uuid, pre_annotation)
                input_result = input_result.combine(pre_anno_result)

        if input_result.status == InputCreationStatus.CREATED:
            input_from_scene_result = _create_input_from_scene(resource, input_result.scene_uuid, project, **kwargs)
            input_result = input_result.combine(input_from_scene_result)

        external_id = scenes_map[input_result.scene_uuid].scene.external_id
        yield input_result.add_external_id(external_id)


def _start_scene_jobs(
    resource: CreateableIOResource,
    scenes: List[Scene],
    dryrun: bool = False,
    feature_flags: Optional[FeatureFlags] = None
) -> List[SceneUuid]:
    # Starts scene jobs. Don't catch validation errors. The returned list must have the same order as the scenes
    return [
        resource.create(scene, dryrun=dryrun, feature_flags=feature_flags).input_uuid for scene in tqdm(scenes, desc='Starting scene jobs')
    ]


def _chunkify(origin_list: list, chunk_size: int) -> List[list]:
    return [origin_list[idx:idx + chunk_size] for idx in range(0, len(origin_list), chunk_size)]


def _wait_for_scene_jobs(
    client: KognicIOClient,
    scene_uuids: List[SceneUuid],
    wait_timeout: float,
    sleep_time: float,
) -> Generator[InputCreationResult, None, None]:

    logger.info(f'Waiting for {len(scene_uuids)} scene jobs to complete')
    remaining_scenes_uuids = set(scene_uuids)
    start_time = time.time()
    chunk_size = DEFAULT_CHUNK_SIZE

    while time.time() - start_time < wait_timeout and remaining_scenes_uuids:
        logger.info(f'{len(remaining_scenes_uuids)} scene jobs remaining')
        for chunk in _chunkify(list(remaining_scenes_uuids), chunk_size):
            try:
                scenes = client.input.get_inputs_by_uuids(chunk)

                # Yield created scenes
                for scene_uuid in set(chunk).difference(set(s.uuid for s in scenes)):
                    yield InputCreationResult(scene_uuid=scene_uuid, status=InputCreationStatus.CREATED)
                    remaining_scenes_uuids.remove(scene_uuid)

                # Yield failed scenes
                for scene in scenes:
                    if scene.status is InputStatus.Failed:
                        yield InputCreationResult(
                            scene_uuid=scene.uuid,
                            status=InputCreationStatus.FAILED,
                            error=InputCreationError(stage=InputCreationStage.SCENE, message=scene.error_message)
                        )
                        remaining_scenes_uuids.remove(scene.uuid)
            except (HTTPError, RuntimeError):
                # Try decreasing the amount of scenes we query the Input API for (only when chunk size is default)
                chunk_size = CONSERVATIVE_CHUNK_SIZE if chunk_size == DEFAULT_CHUNK_SIZE else chunk_size
                break

        time.sleep(sleep_time)

    # Yield results for remaining scenes
    for scene_uuid in remaining_scenes_uuids:
        yield InputCreationResult(scene_uuid=scene_uuid, status=InputCreationStatus.PROCESSING)


def _create_pre_annotation(client: KognicIOClient, scene_uuid: str, pre_annotation: OpenLabelAnnotation) -> InputCreationResult:
    # Try to create a pre-annotation and return a creation result
    try:
        client.pre_annotation.create(scene_uuid, pre_annotation, dryrun=False)
        return InputCreationResult(scene_uuid=scene_uuid, status=InputCreationStatus.CREATED)
    except (HTTPError, RuntimeError) as error:
        logger.error(f"Failed to create pre-annotation for scene {scene_uuid}. {error}")
        creation_error = InputCreationError(stage=InputCreationStage.PRE_ANNOTATION, message=str(error))
        return InputCreationResult(scene_uuid=scene_uuid, status=InputCreationStatus.FAILED, error=creation_error)


def _create_input_from_scene(resource: CreateableIOResource, scene_uuid: SceneUuid, project: str, **kwargs) -> InputCreationResult:
    # Create input from scene. Catch errors and put them in the result
    try:
        resource.create_from_scene(scene_uuid, project, **kwargs)
        return InputCreationResult(scene_uuid=scene_uuid, input_uuid=scene_uuid, status=InputCreationStatus.CREATED)
    except (HTTPError, RuntimeError) as error:
        logger.error(f"Failed to create input from scene {scene_uuid}. {error}")
        creation_error = InputCreationError(stage=InputCreationStage.INPUT_FROM_SCENE, message=str(error))
        return InputCreationResult(scene_uuid=scene_uuid, status=InputCreationStatus.FAILED, error=creation_error)


def _get_creatable_resource(client: KognicIOClient, scenes: List[Scene]) -> CreateableIOResource:
    scene_type = _check_scene_types(scenes)

    resources_map = {
        Cameras: client.cameras,
        CamerasSequence: client.cameras_sequence,
        LidarsAndCameras: client.lidar_and_cameras,
        LidarsAndCamerasSequence: client.lidars_and_cameras_sequence,
        AggregatedLidarsAndCamerasSequence: client.aggregated_lidars_and_cameras_seq
    }

    if scene_type not in resources_map:
        raise ValueError(f"Scene of type '{scene_type.__name__}' not supported")
    return resources_map.get(scene_type)


def _check_scene_types(scenes: List[Scene]) -> Type[Scene]:
    first_type = type(scenes[0])
    assert all([isinstance(scene, first_type) for scene in scenes]), "All scenes must be of the same type"
    return first_type



"""Utility functions for Kognic IO """

from collections.abc import Mapping
from datetime import datetime
from typing import Dict, List

import dateutil.parser


def ts_to_dt(date_string: str) -> datetime:
    """
    Parse string datetime into datetime
    """
    return dateutil.parser.parse(date_string)


def filter_none(js: dict) -> dict:
    if isinstance(js, Mapping):
        return {k: filter_none(v) for k, v in js.items() if v is not None}
    else:
        return js


def get_view_links(input_uuids: List[str]) -> Dict[str, str]:
    """
        For each given input uuid returns an URL where the input can be viewed in the web app.

        :param input_uuids: List with input uuids
        :return Dict: Dictionary mapping each uuid with an URL to view the input.
        """
    view_dict = dict()
    for input_uuid in input_uuids:
        view_dict[input_uuid] = f"https://app.kognic.com/view/input/{input_uuid}"

    return view_dict



from .abstract import *
from .cameras import *
from .cameras_sequence import *
from .input_entry import *
from .input_job import *
from .invalidated_reason_input import *
from .lidars_and_cameras import *
from .lidars_and_cameras_sequence import *
from .aggregated_lidars_and_cameras_seq import *
from .resources import *
from .sensor_specification import *



from typing import Optional, Mapping

from kognic.io.model.input.abstract import BaseInputWithIMUData
from kognic.io.model.input.lidars_and_cameras.frame import Frame
from kognic.io.model.input.metadata.metadata import MetaData
from kognic.io.model.input.resources.resource import Resource
from kognic.io.model.input.sensor_specification import SensorSpecification


class LidarsAndCameras(BaseInputWithIMUData):
    external_id: str
    frame: Frame
    calibration_id: str
    sensor_specification: Optional[SensorSpecification] = None
    metadata: MetaData = dict()

    @property
    def resources(self) -> Mapping[str, Resource]:
        return self.frame.resources



from typing import List, Optional, Mapping

from kognic.io.model.input.abstract.base_input import BaseInput
from kognic.io.model.input.cameras_sequence.frame import Frame
from kognic.io.model.input.metadata.metadata import MetaData
from kognic.io.model.input.resources import Image
from kognic.io.model.input.sensor_specification import SensorSpecification


class CamerasSequence(BaseInput):
    external_id: str
    frames: List[Frame]
    sensor_specification: Optional[SensorSpecification] = None
    metadata: MetaData = dict()

    @property
    def resources(self) -> Mapping[str, Image]:
        mappings = [frame.resources for frame in self.frames]
        superset = {}
        for mapping in mappings:
            superset = {**superset, **mapping}
        return superset



from datetime import datetime

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.util import ts_to_dt


class Project(BaseSerializer):
    created: datetime
    title: str
    description: str
    status: str
    project: str



from __future__ import annotations

import enum
from typing import TypeVar, Optional

from kognic.io.model.input.abstract.base_input import BaseInput
from kognic.openlabel.models import OpenLabelAnnotation
from pydantic import BaseModel

SceneUuid = str
InputUuid = str

Scene = TypeVar("Scene", bound=BaseInput)


class SceneWithPreAnnotation(BaseModel):
    scene: Scene
    pre_annotation: Optional[OpenLabelAnnotation] = None


class InputCreationStage(enum.Enum):
    SCENE = "scene"
    PRE_ANNOTATION = "pre-annotation"
    INPUT_FROM_SCENE = "input-from-scene"


class InputCreationStatus(enum.Enum):
    CREATED = "created"
    FAILED = "failed"
    PROCESSING = "processing"


class InputCreationError(BaseModel):
    stage: InputCreationStage
    message: str


class InputCreationResult(BaseModel):
    scene_uuid: SceneUuid
    status: InputCreationStatus
    external_id: Optional[str] = None
    input_uuid: Optional[InputUuid] = None  # Only defined if an input was created
    error: Optional[InputCreationError] = None  # Only defined if an error occured

    def combine(self, other: InputCreationResult) -> InputCreationResult:
        return InputCreationResult(
            scene_uuid=self.scene_uuid,
            status=other.status,
            external_id=self.external_id or other.external_id,
            input_uuid=self.input_uuid or other.input_uuid,
            error=self.error or other.error,
        )

    def add_external_id(self, external_id: str) -> InputCreationResult:
        return InputCreationResult(
            scene_uuid=self.scene_uuid,
            status=self.status,
            external_id=external_id,
            input_uuid=self.input_uuid,
            error=self.error,
        )



from typing import Dict

from pydantic import Field

from kognic.io.model.base_serializer import BaseSerializer


class InputJobCreated(BaseSerializer):
    input_uuid: str = Field(alias='internalId')
    files: Dict[str, str]

    def __str__(self):
        return f"{self.__class__.__name__}(input_uuid={self.input_uuid}, files={{...}})"


class CreateInputResponse(BaseSerializer):
    input_uuid: str

    @staticmethod
    def from_input_job_response(input_job: InputJobCreated):
        return CreateInputResponse(input_uuid=input_job.input_uuid)



from abc import ABC

from kognic.io.model.input.abstract.base_frame import BaseFrame
from kognic.io.model.input.metadata.metadata import FrameMetaData


class SequenceFrame(BaseFrame, ABC):
    frame_id: str
    relative_timestamp: int
    metadata: FrameMetaData = dict()



from abc import ABC
from typing import Mapping

from kognic.io.model import BaseSerializer
from kognic.io.model.input.resources.resource import Resource


class BaseFrame(BaseSerializer, ABC):
    """
    Base class for the concept of a frame of data in sequential scenes.
    """

    @property
    def resources(self) -> Mapping[str, Resource]:
        raise NotImplementedError(f"resources not implemented for {type(self)}")



from typing import Dict

from pydantic import Field

from kognic.io.model.base_serializer import BaseSerializer


class UploadUrls(BaseSerializer):
    files_to_url: Dict[str, str] = Field(alias="files")
    input_uuid: int = Field(alias="jobId")



from typing import Mapping

from kognic.io.model import PointCloud
from kognic.io.model.input.abstract import BaseInputWithIMUData
from kognic.io.model.input.lidars.frame import Frame
from kognic.io.model.input.metadata.metadata import MetaData


class Lidars(BaseInputWithIMUData):
    external_id: str
    frame: Frame
    metadata: MetaData = dict()

    @property
    def resources(self) -> Mapping[str, PointCloud]:
        return self.frame.resources



from typing import List, Mapping

from kognic.io.model.input.abstract.sequence_frame import SequenceFrame
from kognic.io.model.input.resources import Image, VideoFrame


class Frame(SequenceFrame):
    images: List[Image] = []
    video_frames: List[VideoFrame] = []

    @property
    def resources(self) -> Mapping[str, Image]:
        # Video is not included as of 2022-10; currently unused and unsupported.
        return {i.resource_id: i for i in self.images}



from kognic.io.model.calibration.calib import SensorCalibration, SensorCalibrationEntry
from kognic.io.model.calibration.common import CalibrationType, RotationQuaternion, Position, BaseCalibration
from kognic.io.model.calibration.camera import *
from kognic.io.model.calibration.lidar import *



from datetime import datetime
from enum import Enum

from kognic.io.model.base_serializer import BaseSerializer


class ProjectBatchStatus(str, Enum):
    PENDING = 'pending'
    OPEN = 'open'
    READY = 'ready'
    INPROGESS = 'in-progress'
    COMPLETED = 'completed'
    ARCHIVED = 'archived'


class ProjectBatch(BaseSerializer):
    project: str
    batch: str
    title: str
    status: ProjectBatchStatus
    created: datetime
    updated: datetime




from typing import List, Optional, Mapping

from kognic.io.model.input.abstract import BaseInputWithIMUData
from kognic.io.model.input.aggregated_lidars_and_cameras_seq.frame import Frame
from kognic.io.model.input.metadata.metadata import MetaData
from kognic.io.model.input.resources.resource import Resource
from kognic.io.model.input.sensor_specification import SensorSpecification


class AggregatedLidarsAndCamerasSequence(BaseInputWithIMUData):
    external_id: str
    frames: List[Frame]
    calibration_id: str
    sensor_specification: Optional[SensorSpecification] = None
    metadata: MetaData = dict()

    @property
    def resources(self) -> Mapping[str, Resource]:
        mappings = [frame.resources for frame in self.frames]
        superset = {}
        for mapping in mappings:
            superset = {**superset, **mapping}
        return superset



import os
from pprint import pprint

import click
from kognic.io.client import KognicIOClient
from kognic.io.util import get_view_links
from tabulate import tabulate


def _tabulate(body, headers, title=None):
    tab = tabulate(
        body,
        headers=headers,
        tablefmt='fancy_grid',
    )
    if title:
        title_len = len(title)
        spacing = len(tab.split('\n')[0])

        roof = "=" * spacing
        floor = "=" * spacing

        left_multi = spacing//2 - title_len//2 - 1

        title_text_left = " "*left_multi + title
        title_text = title_text_left

        title_house = roof + '\n' + title_text + '\n' + floor
        tab = title_house + '\n' + tab
    return tab


def _get_table(sequence, headers, title=None):
    body = []
    for p in sequence:
        body.append([vars(p)[h] for h in headers])
    return _tabulate(body, headers, title)


@click.group()
@click.pass_context
def cli(ctx):
    """A CLI wrapper for Kognic utilities"""


@click.command()
@click.argument('project', nargs=1, default=None, required=False, type=str)
@click.option('--get-batches', is_flag=True)
@click.pass_obj
def projects(obj, project, get_batches):
    client = obj['client']
    print()
    if project and get_batches:
        list_of_input_batches = client.project.get_project_batches(project)
        headers = ["created", "project", "batch", "title", "status", "updated"]
        tab = _get_table(list_of_input_batches, headers, "BATCHES")
        print(tab)
    elif project:
        list_of_projects = client.project.get_projects()
        target_project = [p for p in list_of_projects if p.project == project]
        headers = ["created", "project", "title", "description", "status"]
        tab = _get_table(target_project, headers, "PROJECTS")
        print(tab)
    else:
        list_of_projects = client.project.get_projects()
        headers = ["created", "project", "title", "description", "status"]
        tab = _get_table(list_of_projects, headers, "PROJECTS")
        print(tab)


@click.command()
@click.option('--project', nargs=1, default=None, required=False, type=str)
@click.option('--batch', nargs=1, default=None, required=False, type=str)
@click.option('--external-ids', required=False, multiple=True)
@click.option('--include-invalidated', required=False, is_flag=True)
@click.option("--view", is_flag=True)
@click.option("--uuids", type=str, help="Comma separated list of input uuids")
@click.pass_obj
def inputs(obj, project, batch, external_ids, include_invalidated, view, uuids):
    client = obj['client']
    print()
    if view and project:
        inputs = client.input.get_inputs(project, batch, external_ids=external_ids, include_invalidated=include_invalidated)
        view_dict = {input.uuid: input.view_link for input in inputs}
        body = []
        headers = ["uuid", "view_link"]
        for uuid, link in view_dict.items():
            body.append([uuid, link])
        tab = _tabulate(body, headers, title="VIEW LINKS FOR INPUTS")
        print(tab)
    elif uuids is not None:
        uuids = uuids.split(",")
        inputs = client.input.get_inputs_by_uuids(uuids)
        headers = ["uuid", "external_id", "batch", "input_type", "status", "view_link", "error_message"]
        tab = _get_table(inputs, headers, "INPUTS")
        print(tab)
    else:
        inputs = client.input.get_inputs(project, batch, include_invalidated=include_invalidated)
        headers = ["uuid", "external_id", "batch", "input_type", "status", "view_link", "error_message"]
        tab = _get_table(inputs, headers, "INPUTS")
        print(tab)


@click.command()
@click.argument('input_uuid', nargs=1, required=True, type=str)
@click.pass_obj
def view(obj, input_uuid):
    client = obj['client']
    print()
    view_dict = get_view_links([input_uuid])
    body = [[input_uuid, view_dict[input_uuid]]]
    headers = ["uuid", "view_link"]
    tab = _tabulate(body, headers, title="VIEW LINK")
    print(tab)


@click.command()
@click.option('--id', nargs=1, default=None, required=False, type=str)
@click.option("--external-id", nargs=1, required=False, type=str)
@click.option('--raw', is_flag=True)
@click.pass_obj
def calibration(obj, id, external_id, raw):
    client = obj['client']
    print()
    if id is not None:
        calibration_entry = client.calibration.get_calibration(id)
        headers = ["id", "external_id", "created"]
        tab = _get_table([calibration_entry], headers, "CALIBRATION")
        print(tab)
        if raw:
            print()
            pprint(calibration_entry.calibration)
    elif external_id is not None:
        list_of_calibrations = client.calibration.get_calibrations(external_id=external_id)
        headers = ["id", "external_id", "created"]
        tab = _get_table(list_of_calibrations, headers, "CALIBRATION")
        print(tab)
    else:
        list_of_calibrations = client.calibration.get_calibrations()
        headers = ["id", "external_id", "created"]
        tab = _get_table(list_of_calibrations, headers, "CALIBRATION")
        print(tab)


cli.add_command(projects)
cli.add_command(inputs)
cli.add_command(calibration)
cli.add_command(view)


def main():
    env = os.getenv("KOGNIC_CLIENT_ORGANIZATION_ID", None)
    if env:
        org_id = int(env)
        print("<" * 25, f" Acting on behalf of organization {org_id}", 25 * ">")
    else:
        org_id = None

    client = KognicIOClient(auth=None, client_organization_id=org_id)

    cli(obj={'client': client}, prog_name="kognicutil")



from typing import Dict, List

from kognic.io.model.ego.imu_data import IMUData
from kognic.io.model.input.abstract.base_input import BaseInput


class BaseInputWithIMUData(BaseInput):
    """
    Base class for any input that has IMU data.
    """
    imu_data: List[IMUData] = []

    def to_dict(self, by_alias=True) -> Dict:
        # IMU data is excluded as it's posted separately.
        return self.dict(exclude_none=True,
                         by_alias=by_alias,
                         exclude={'imu_data': True})



from typing import Optional

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration.common import BaseCalibration, CalibrationType


class LidarFieldOfView(BaseSerializer):
    start_angle_deg: float
    stop_angle_deg: float
    depth: Optional[float] = None


class LidarCalibration(BaseCalibration):
    calibration_type = CalibrationType.LIDAR
    field_of_view: Optional[LidarFieldOfView] = None



from abc import ABC

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.input.resources.resource import Resource

camera_sensor_default = "CAM"


class VideoTS(ABC, BaseSerializer):
    video_timestamp: int


class VideoFrame(Resource, VideoTS):
    sensor_name: str = camera_sensor_default
    video_timestamp: int



from kognic.io.model.calibration.camera.common import BaseCameraCalibration
from kognic.io.model.calibration.common import CalibrationType
from kognic.io.model.base_serializer import BaseSerializer


class KannalaDistortionCoefficients(BaseSerializer):
    k1: float
    k2: float
    p1: float
    p2: float


class UndistortionCoefficients(BaseSerializer):
    l1: float
    l2: float
    l3: float
    l4: float


class KannalaCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.KANNALA.value
    distortion_coefficients: KannalaDistortionCoefficients
    undistortion_coefficients: UndistortionCoefficients



import mimetypes
from enum import Enum
from io import IOBase
from typing import Optional, Callable, BinaryIO, Union

FileDataContent = Union[bytes, BinaryIO, IOBase]
FileDataCallback = Callable[[str], FileDataContent]


class FileData:
    """
    Describes a source of file data to upload at a later point.
    """

    class Format(str, Enum):
        """
        Enumeration of file data formats supported as input resources.
        """
        CSV = "csv"
        LAS = "las"
        LAZ = "laz"
        PCD = "pcd"
        JPG = "jpg"
        PNG = "png"

        @property
        def is_pointcloud(self):
            return self in (FileData.Format.CSV, FileData.Format.LAS, FileData.Format.LAZ, FileData.Format.PCD)

        @property
        def is_image(self):
            return self in (FileData.Format.JPG, FileData.Format.PNG)

    format: Format
    data: Optional[FileDataContent]
    callback: Optional[FileDataCallback]

    def __init__(self, format: Format, data: Optional[FileDataContent] = None,
                 callback: Optional[FileDataCallback] = None):
        """
        Create a new FileData, describing the content of a file. The data in the file is taken from either a blob
        provided directly at creation time, or from a callback that will be called at some later point.

        :param format: The format of the data in this file. Must be one of the supported types in FileData.Format.
        :param data: Optional, the data itself. May be bytes or a bytes-like object.
        :param callback: Optional, a callback which can provide the data later, as bytes or a bytes-like object.
        """
        self.format = format
        self.data = data
        self.callback = callback
        self.check_data_source()

    def check_data_source(self):
        if self.data is None and self.callback is None:
            raise ValueError(f"One of data or callback must be set on FileData")
        if self.data is not None and self.callback is not None:
            raise ValueError(f"Only one of data or callback may be set on FileData")

    @property
    def suffix(self):
        return f'.{self.format.value}'

    @property
    def content_type(self):
        if self.suffix in mimetypes.types_map:
            return mimetypes.types_map[self.suffix]
        else:
            return 'application/octet-stream'

    def __repr__(self):
        data_part = '' if self.data is None else 'blob'
        callback_part = '' if self.callback is None else 'callback'
        return f"FileData({data_part}{callback_part} format={self.format})"



from typing import List, Mapping, Union, Optional

from kognic.io.model import UnixTimestampNs
from kognic.io.model.ego import EgoVehiclePose
from kognic.io.model.input.abstract.sequence_frame import SequenceFrame
from kognic.io.model.input.resources import PointCloud


class Frame(SequenceFrame):
    point_clouds: List[PointCloud]
    ego_vehicle_pose: Optional[EgoVehiclePose] = None
    unix_timestamp: Optional[UnixTimestampNs] = None

    @property
    def resources(self) -> Mapping[str, PointCloud]:
        return {p.resource_id: p for p in self.point_clouds}



from typing import Optional, Generator

from kognic.io.model.annotation.client_annotation import Annotation, PartialAnnotation
from kognic.io.resources.abstract import IOResource


class AnnotationResource(IOResource):

    def get_project_annotations(self,
                                project: str,
                                annotation_type: str,
                                batch: Optional[str] = None,
                                include_content: bool = True,
                                ) -> Generator[Annotation, None, None]:
        """
        Gets annotations for a project and annotation type with an optional filter for a batch.
        Annotations include annotation contents which means that this function is heavier than `get_partial_annotations`
        that only gets partial annotations.

        :param project: project to query annotations from
        :param annotation_type: annotation type to query annotation on
        :param batch: batch to filter on (optional)
        :param include_content: whether to include annotation content (default: True)
        :returns Generator[Annotation]: Generator of annotations
        """

        url = f"v1/annotations/projects/{project}/"
        if batch:
            url += f"batch/{batch}/"

        url += f"annotation-type/{annotation_type}/search"

        for js in self._paginate_get(url):
            partial_annotation = PartialAnnotation.from_json(js)
            content = self._file_client.get_json(partial_annotation.uri) if include_content else None
            yield partial_annotation.to_annotation(content)

    def get_annotation(self, input_uuid: str, annotation_type: str) -> Annotation:
        """
        Gets an annotations for an input and annotation type. A NotFound exception will be raised if there isn't any
        annotation for the given uuid and annotation type. Use `client.input.get_inputs_by_uuids` to find out whether
        the input has been configured for the annotation type or not.

        :param input_uuid: uuid of the input
        :param annotation_type: annotation type to get annotation for
        :returns Annotation: annotation with its content
        """
        json_resp = self._client.get(f"v1/annotations/inputs/{input_uuid}/annotation-type/{annotation_type}")
        annotation = Annotation.from_json(json_resp)
        return annotation
